;;  Parallel fft with separate shuffles                  File: fft2.lisp

(in-package '*lisp)

;; fftshufl.lisp should be loaded before this.  It initializes pvars
;;
;;	c0	grid coordinate 0
;;	cr	bit-reversal of c0
;;	c1	grid coordinate 1
;;	cs	cube address for shuffle on axis 0

;; Calculate N-point fft's along axis 0 of current 2-D geometry.
;; N is the dimension in axis 0.  Does no shuffles.  Address bits
;; processed high to low if ud = DN, low to high if UP.
;; Standard DFT is calculated if (not inv), unnormalized inverse if inv.
;;
;; The use of axis 0 in a 2-D geometry is simply to make things
;; definite.

(defvar test nil)    ;; Controls intermediate printing

(*defun Ufftx (Z ud inv)
  (declare (type single-complex-pvar Z))
  (*let ((Z1 (!! #C(0.0 0.0)))  (Z2 (!! #C(0.0 0.0)))  (W  (!! #C(0.0 0.0)))
	 (dx (!! 0)) (jr (!! 0)))
	(declare (type single-complex-pvar Z1 Z2 W))
	(declare (type (field-pvar csz) dx jr))
     (let* ((N (dimension-size 0)) (theta 0.0)
	    (Nx 1) (N1 1) (N2 2) (Np 0) (Nn (logcount (1- N))))
           (declare (type fixnum N Nx N1 N2 Np Nn))
           (declare (type single-float theta))
       ;; Adjust for up/dn, nrm/inv
       (if (eq ud 'DN)
	   (progn (setq Nx -1 N1 (ash N -1) Np (1- Nn))
		  (*set jr cr))
	   (*set jr c0))         ;; jr is effective coordinate
       (when test (print (list N Nx N1 Np)) (ppp jr))
       (*set Z2 Z)
       (when (> N 1)	;; Non-trivial problem
	;; Combine pairs
        ;; Send values to "other" processors
        (*set dx (logxor!! c0 (!! N1)))
        (*pset :no-collisions Z2 Z1 (cube-from-grid-address!! dx c1))
        (*if (logbitp!! (!! Np) c0)  (*set Z2 (-!! Z1 Z2))
	                             (*set Z2 (+!! Z2 Z1)))
        (when test (ppp Z2))
	;; Loop to combine larger groups
	;; Z2 = result of previous combining steps
	(dotimes (k (1- Nn))
	  (setq N1 (ash N1 Nx) Np (+ Np Nx) N2 (ash N2 1))
	  (*set dx (logxor!! c0 (!! N1)))
          (setq theta (/ (* 2.0 pi) N2))
          (if inv (setq theta (- theta)))       ;; -theta = 1/theta
	  (when test (print (list k N1 Np N2 theta)))
	  ;; Send values to "other" processors
	  (*pset :no-collisions Z2 Z1 (cube-from-grid-address!! dx c1))
	  ;; Swap where needed so Z2 = even entry, Z1 = odd entry
	  (*when (logbitp!! (!! Np) c0)
	     (*set W Z2) (*set Z2 Z1) (*set Z1 W))
          (*set W (cis!! (*!! (mod!! jr (!! N2)) theta)))  ;; w^j, so to speak
	  (when test (ppp W))
          ;; Form result in Z2
          (*set Z2 (+!! Z2 (*!! W Z1)))
          (when test (ppp Z2))
         )) ;; end dotimes, end when (> N 1)
       Z2
     )  ;; end let*
   )  ;; end *let
)

;; Polynomial multiplication test (see fftest.lisp)

(defun ptest2 ()
  (poly)
  (*set xt (Ufftx xx 'DN nil))
  (*set yt (Ufftx yy 'DN nil))
  (*set zt (*!! xt yt))
  (*set zz (/!! (Ufftx zt 'UP t) (!! 8.0)))
  (ppp zz)
)
